/*
 *  linux/fs/qnx4fs/common.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 *  14 jan 2004 Dmitry Yakovlev - 2.6 kernel support
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/fs.h>
#include "qnx4fs_fs.h"
#include <linux/buffer_head.h>

struct buffer_head *
qnx4fs_get_slot(struct inode *ip, ino_t *inop, qnx4fs_dirent_t **qdpp, int n)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	unsigned long phys;
	long nblocks, block;
	int slot = 1;

	nblocks = qnx4fs_nblocks(ip);
	for (block = 0; block < nblocks; block++, slot = 0) {
		phys = qnx4fs_map_block(ip, block);
		if (!phys)
			return ERR_PTR(-ENOSPC);

		bp = qnx4fs_sb_bread(ip->i_sb, phys - 1);
		if (bp == NULL) {
			QNX4FS_WARN(" can't read block %lu, ino %ld\n", phys,
				 ip->i_ino);
			return ERR_PTR(-ENOSPC);
		}
		qdp = (qnx4fs_dirent_t *)bp->b_data + slot;
		for (; slot < QNX4FS_DIRENTS_PER_BLOCK; slot++, qdp++) {
			if (qdp->i.status == 0) {
				*inop = direct2ino(phys, slot);
				*qdpp = qdp;
				goto found;
			}
		}
		brelse(bp);
	}
	bp = qnx4fs_iexpand_special(ip, n);
	if (bp == NULL)
		return ERR_PTR(-ENOSPC);
	*inop = direct2ino(bp->b_blocknr + 1, 0);
	*qdpp = (qnx4fs_dirent_t *)bp->b_data;
found:
	return bp;
}

struct buffer_head *
qnx4fs_sb_bread(struct super_block *sb, unsigned long block)
{
	struct buffer_head *bh;

	if (block >= QNX4FS_SB(sb)->s_blocks)
		goto wrong_block;

	if ((block < (QNX4FS_SB(sb)->s_bitmap_blocks + QNX4FS_BITMAP_BLOCK - 1))
	&& (block != (QNX4FS_ROOT_BLOCK - 1)))
		goto wrong_block;

	bh = sb_bread(sb, block);
	if (!bh)
		QNX4FS_WARN(" unable to read block %lu\n", block);
	return	bh;

wrong_block:
	QNX4FS_WARN(" wrong block number %lu\n", block);
	return NULL;
}

void
qnx4fs_mark_slot(qnx4fs_dirent_t *qdp, struct buffer_head *bp, int status)
{
	if (!status)
		qdp->i.nlink = 0;
	qdp->i.status = status;
	mark_buffer_dirty(bp);
	brelse(bp);
}
