/*
 *  linux/fs/qnx4fs/dir.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 *  14 jan 2004 Dmitry Yakovlev - 2.6 kernel support
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/fs.h>
#include "qnx4fs_fs.h"
#include <linux/buffer_head.h>
#include <linux/smp_lock.h>

static int
qnx4fs_readdir(struct file *fp, void *dirent, filldir_t filldir)
{
	struct inode *dip = fp->f_dentry->d_inode;
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	unsigned long phys;
	long nblocks, block;
	int slot, namelen;
	ino_t ino;

	/* mutex_lock(&QNX4FS_SB(dip->i_sb)->s_slock); */

	nblocks = qnx4fs_nblocks(dip);
	block = fp->f_pos >> QNX4FS_DIRENTS_PER_BLOCK_BITS;
	slot = fp->f_pos & (QNX4FS_DIRENTS_PER_BLOCK - 1);

	QNX4FS_DEBUG(" inode %ld, block %d, slot %d\n", dip->i_ino,
		(int)block, slot);

	for (; block < nblocks; block++, slot = 0) {
		phys = qnx4fs_map_block(dip, block);
		if (!phys) {
			QNX4FS_WARN(" can't map block %ld\n", block);
			goto out;
		}
		bp = qnx4fs_sb_bread(dip->i_sb, phys - 1);
		if (bp == NULL) {
			QNX4FS_WARN(" can't read block %lu\n", phys);
			goto out;
		}
		qdp = (qnx4fs_dirent_t *)bp->b_data + slot;
		for (; slot < QNX4FS_DIRENTS_PER_BLOCK; slot++, qdp++) {
			if (qdp->i.status & QNX4FS_FILE_LINK) {
				namelen = QNX4FS_LONG_NAME_MAX;
				ino = link2ino(qdp);
			} else if (qdp->i.status & QNX4FS_FILE_USED) {
				namelen = QNX4FS_SHORT_NAME_MAX;
				ino = direct2ino(phys, slot);
			} else {
				continue;
			}

			if (ino == QNX4FS_INODES_INO
			    || ino == QNX4FS_SB(dip->i_sb)->s_bitmap_ino) {
				continue;
			}
			namelen = strnlen(qdp->i.name, namelen);
			if (filldir(dirent, qdp->i.name, namelen,
				(block << QNX4FS_DIRENTS_PER_BLOCK_BITS) | slot,
				ino, DT_UNKNOWN) < 0) {

				brelse(bp);
				goto out;
			}
		}
		brelse(bp);
	}
out:

	fp->f_pos = (block << QNX4FS_DIRENTS_PER_BLOCK_BITS) | slot;
	fp->f_version = dip->i_version;
	/* mutex_unlock(&QNX4FS_SB(dip->i_sb)->s_slock); */

	return 0;
}

const struct file_operations qnx4fs_dir_ops = {
	.read		= generic_read_dir,
	.fsync		= file_fsync,
	.readdir	= qnx4fs_readdir,
};
